/* mmgetaa.c - functions for mmap()ed access to libraries */

/* copyright (c) 1999,2000 William R. Pearson */

/* version 0 September, 1999 */

/*
  This is one of three alternative files that can be used to
  read a database.  The three files are nxgetaa.c, nmgetaa.c, and
  mmgetaa.c.

  nxgetaa.c contains the original code for reading databases, and
  is still used for Mac and PC versions of fasta33 (which do not
  use mmap).

  nmgetaa.c and mmgetaa.c are used together.  nmgetaa.c provides
  the same functions as nxgetaa.c if memory mapping is not used,
  mget IBM_AIXaa.c provides the database reading functions if memory
  mapping is used. The decision to use memory mapping is made on
  a file-by-file basis.
*/

/* $Name: fa33t09d0 $ - $Id: mmgetaa.c,v 1.12.2.2 2001/03/27 19:42:25 wrp Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#ifdef IBM_AIX
#include <fcntl.h>
#else
#include <sys/fcntl.h>
#endif

#define XTERNAL
#include "uascii.h"
#include "upam.h"
#undef XTERNAL

extern int nsfnum;	/* number of superfamily numbers */
extern int sfnum[10];	/* superfamily number from types 0 and 5 */
extern int nsfnum_n;
extern int sfnum_n[10];

#ifdef SUPERFAMNUM
static char tline[MAXLINE];
#endif

#define GCGBIN 6

#ifndef MAP_FILE
#define MAP_FILE 0
#endif

extern long bl2_long8_cvt(long);



#include "mm_file.h"
static struct m_file_str *m_fptr, m_file;

static int lpos;

/* load_mmap() loads the d_pos[] and s_pos[] arrays for rapid access */

struct m_file_str *
load_mmap(FILE *libi,	/* fd for already open ".xin" file */
	  char *sname,	/* name of sequence database file */
	  int lib_type,	/* 0-Fasta, 5-vms_pir, 6-gcg_binary */
	  int ldnaseq)	/* 1 for DNA, 0 for protein */
{
  char format[4];
  int i;
  long f_size;
  struct stat statbuf;
  int mm_flag, mm64_flag;
  int *tmp_pos_arr;

  /* first check that the necessary indices are up-to-date */
  /* read the offsets in ".xin" file */
  if (fread(format,1,4,libi)==0) {
    fprintf(stderr," cannot read .xin format\n");
    return NULL;
  }
    
  mm64_flag = (format[2]==1);	/* 4 bytes or 8 bytes for long? */

  if (format[3]!=lib_type) {
    fprintf(stderr," cannot read format %d != lib_type %d\n",
	    format[3],lib_type);
    return NULL;
  }

  src_int4_read(libi,&m_file.lib_aa);
  if (m_file.lib_aa == ldnaseq) { /* database residue mismatch */
    fprintf(stderr," residue type mismatch %s != %s (.xin) in %s\n",
	    (m_file.lib_aa ? "DNA" : "prot."),(ldnaseq ? "prot." : "DNA"),
	    sname);
    return NULL;
  }
    
  /* get get file size from index */
  if (mm64_flag) src_long8_read(libi,&f_size);
  else src_long4_read(libi,&f_size);

  /* now, start to open mmap()ed file */
  mm_flag=((m_file.mmap_fd=open(sname,O_RDONLY))>=0);
  if (!mm_flag) {
    fprintf(stderr," cannot open %s for mmap()", sname);
    perror("...");
    return NULL;	/* file did not open */
  }

  /* fstat the library file and get size */
  if(fstat(m_file.mmap_fd, &statbuf) < 0) {
    fprintf(stderr," cannot stat %s for mmap()", sname);
    perror("...");
    mm_flag = 0;
    goto finish;
  }

  /* check for identical sizes - if different, do not mmap */
  if (f_size != statbuf.st_size) {
    fprintf(stderr," %s file size (%ld) and expected size (%ld) don't match\n",
	    sname,statbuf.st_size,f_size);
    mm_flag = 0;
    goto finish;    
  }

  /* the the index file and library file are open and the sizes match */
  /* allocate the m_file struct and  map the file */

  if ((m_fptr = calloc(1,sizeof(struct m_file_str)))==NULL) {
    fprintf(stderr," cannot allocate m_file_str\n");
    mm_flag = 0;
    goto finish;
  }
  m_file.st_size = statbuf.st_size;
  if((m_file.mmap_base = 
      mmap(NULL, m_file.st_size, PROT_READ,
	   MAP_FILE | MAP_SHARED, m_file.mmap_fd, 0)) == (char *) -1) {
    mm_flag = 0;
    fprintf(stderr," cannot mmap %s", sname);
    perror("...");
  }  
 finish:
  close(m_file.mmap_fd);
  if (!mm_flag) { return NULL; }

  /* now finish reading the index file */
  src_int4_read(libi,&m_file.max_cnt);
  if (mm64_flag) {
    src_long8_read(libi,&m_file.tot_len);
    src_long8_read(libi,&m_file.max_len);
  }
  else {
    src_long4_read(libi,&m_file.tot_len);
    src_long4_read(libi,&m_file.max_len);
  }

#ifdef DEBUG
  fprintf(stderr,
	  "%s\tformat: %c%c%d %d; max_cnt: %d; tot_len: %ld max_len: %ld\n",
	  sname,format[0],format[1],format[2],format[3],
	  m_file.max_cnt,m_file.tot_len,m_file.max_len);
#endif

  /* allocate array of description pointers */
  if (!mm64_flag) {
    if ((tmp_pos_arr=(int *)calloc(m_file.max_cnt+1,sizeof(int)))==NULL) {
      fprintf(stderr," cannot allocate %d for tmp_pos array\n",
	      m_file.max_cnt+1);
    }
  }

  if ((m_file.d_pos_arr=(unsigned long *)calloc(m_file.max_cnt+1, sizeof(long)))==NULL) {
    fprintf(stderr," cannot allocate %d for desc. array\n",m_file.max_cnt+1);
    exit(1);
  }
  /* read m_file.d_pos[m_file.max_cnt+1] */
  if (mm64_flag) {
    if (fread(m_file.d_pos_arr,sizeof(long),m_file.max_cnt+1,libi)!=
	m_file.max_cnt+1) {
      fprintf(stderr," error reading desc. offsets: %s\n",sname);
      return NULL;
    }
  }
  else {
    if (fread(tmp_pos_arr,sizeof(int),m_file.max_cnt+1,libi)!=
	m_file.max_cnt+1) {
      fprintf(stderr," error reading desc. offsets: %s\n",sname);
      return NULL;
    }
  }

#ifndef IS_BIG_ENDIAN
  if (mm64_flag)
    for (i=0; i<=m_file.max_cnt; i++)
      m_file.d_pos_arr[i] = bl2_long8_cvt(m_file.d_pos_arr[i]);
  else
    for (i=0; i<=m_file.max_cnt; i++)
      m_file.d_pos_arr[i] = bl2_uint4_cvt(tmp_pos_arr[i]);
#else
  if (!mm64_flag) 
    for (i=0; i<=m_file.max_cnt; i++)
      m_file.d_pos_arr[i] = tmp_pos_arr[i];
#endif

  /* allocate array of sequence pointers */
  if ((m_file.s_pos_arr=(unsigned long *)calloc(m_file.max_cnt+1,sizeof(long)))==NULL) {
    fprintf(stderr," cannot allocate %d for seq. array\n",m_file.max_cnt+1);
    exit(1);
  }

  /* read m_file.s_pos[m_file.max_cnt+1] */
  if (mm64_flag) {
    if (fread(m_file.s_pos_arr,sizeof(long),m_file.max_cnt+1,libi)!=
	m_file.max_cnt+1) {
      fprintf(stderr," error reading seq offsets: %s\n",sname);
      return NULL;
    }
  }
  else {
    if (fread(tmp_pos_arr,sizeof(int),m_file.max_cnt+1,libi)!=
	m_file.max_cnt+1) {
      fprintf(stderr," error reading seq offsets: %s\n",sname);
      return NULL;
    }
  }

#ifndef IS_BIG_ENDIAN
  if (mm64_flag)
    for (i=0; i<=m_file.max_cnt; i++)
      m_file.s_pos_arr[i] = bl2_long8_cvt(m_file.s_pos_arr[i]);
  else
    for (i=0; i<=m_file.max_cnt; i++)
      m_file.s_pos_arr[i] = bl2_uint4_cvt(tmp_pos_arr[i]);
#else
  if (!mm64_flag) 
    for (i=0; i<=m_file.max_cnt; i++)
      m_file.s_pos_arr[i] = tmp_pos_arr[i];
#endif

  if (!mm64_flag) free(tmp_pos_arr);

  memcpy(m_fptr,&m_file,sizeof(struct m_file_str));

  lpos = 0;
  fclose(libi);
  return m_fptr;
}  
 
void
reload_mmap(struct m_file_str *m_fptr) {
  memcpy(&m_file,m_fptr,sizeof(struct m_file_str));
  lpos = 0;
}

char *mgets (char *s, int n)
{
  char *cs, *mfp;

  mfp = m_file.mmap_addr;
  cs = s;

  while (--n > 0 && (*mfp != (char)EOF))
    if ((*cs++ = *mfp++) == '\n') break;
  *cs = '\0';

  m_file.mmap_addr = mfp;
  return (*mfp == (char)EOF && cs == s) ? NULL : s;
}

agetlibm(seq,maxs,libstr,libpos,lcont)
     unsigned char *seq;
     int maxs;
     char *libstr;
     long *libpos;
     int *lcont;
{
  register unsigned char *cp, *seqp;
  register int *ap;
  unsigned char *seqm, *seqm1;
  char *bp;
  static long seq_len;
  static unsigned char *cp_max;
#ifdef SUPERFAMNUM
  char *bp1, *bpa, *tp;
  int i;
#endif

  seqp = seq;
  seqm = &seq[maxs-9];
  seqm1 = seqm-1;

#if !defined(TFAST)
  ap = sascii;
#else
  ap = nascii;
#endif

  if (*lcont==0) {
    if (lpos >= m_file.max_cnt) return (-1);
    seq_len = m_file.d_pos_arr[lpos+1] - m_file.s_pos_arr[lpos];
    *libpos = lpos;

#ifndef PCOMPLIB
    strncpy(libstr,m_file.mmap_base+m_file.d_pos_arr[lpos]+1,12);
    libstr[12]='\0';
    if ((bp=strchr(libstr,' '))!=NULL) *bp='\0';
#else
    strncpy(libstr,m_file.mmap_base+m_file.d_pos_arr[lpos]+1,255);
    libstr[255]='\0';
#endif
    if ((bp=strchr(libstr,'\n'))!=NULL) *bp='\0';

#ifdef SUPERFAMNUM
    sfnum[0]=nsfnum=0;
    strncpy(tline,m_file.mmap_base+m_file.d_pos_arr[lpos]+1,sizeof(tline));
    tline[MAXLINE-1]='\0';
    if ((bp=strchr(tline,'\n'))!=NULL) *bp='\0';
    if ((bp=strchr(tline,' ')) && (bp=strchr(bp+1,SFCHAR))) {
      if ((bpa = strchr(bp+1,'\001'))!=NULL) *bpa = '\0';
      if ((bp1=strchr(bp+1,SFCHAR))==NULL) {
	fprintf(stderr," second %c missing: %s\n",SFCHAR,tline);
      }
      else {
	*bp1 = '\0';
	i = 0;
	if ((tp = strtok(bp+1," \t"))!=NULL) {
	  sfnum[i++] = atoi(tp);
	  while ((tp = strtok((char *)NULL," \t")) != (char *)NULL) {
	    sfnum[i++] = atoi(tp);
	    if (i>=9) break;
	  }
	}
	sfnum[nsfnum=i]= 0;
	if (nsfnum>1) sf_sort(sfnum,nsfnum);
	else {
	  if (nsfnum<1) fprintf(stderr," found | but no sfnum: %s\n",libstr);
	}
      }
    }
    else {
      sfnum[0] = nsfnum = 0;
      }
#else
    sfnum[0]=nsfnum = 0;
#endif

    m_file.mmap_addr = m_file.mmap_base+m_file.s_pos_arr[lpos];
    cp_max = (unsigned char *)(m_file.mmap_addr+seq_len);
  }

  for (cp=(unsigned char *)m_file.mmap_addr; seqp<seqm1; ) {
    if ((*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA) continue;
    --seqp;
    if (cp >= cp_max) break;
  }
  m_file.mmap_addr = (char *)cp;

  if (seqp>=seqm1) (*lcont)++;
  else {
    *lcont=0;
    lpos++;
  }

  *seqp = EOSEQ;
  /*   if ((int)(seqp-seq)==0) return 1; */
  return (int)(seqp-seq);
}

void
aranlibm(str,cnt,libpos)
	char *str; int cnt;
	long libpos;
{
  char *bp;
  int llen;

  llen = m_file.s_pos_arr[libpos]-m_file.d_pos_arr[libpos];
  if (llen >= cnt) llen = cnt-1;

  strncpy(str,m_file.mmap_base+m_file.d_pos_arr[libpos]+1,llen-1);
  str[llen-1]='\0';
  while ((bp = strchr(str,'\001'))!=NULL ||
	 (bp=strchr(str,'\t'))!=NULL) *bp=' ';
  if ((bp = strchr(str,'\r'))!=NULL) *bp='\0';
  if ((bp = strchr(str,'\n'))!=NULL) *bp='\0';
  lpos = libpos;
}

/* there is no vgetlibm() because vgetlibm() and agetlibm() are
   identical - the difference in the two file formats relates to the
   location of the sequence, which is already available in spos_arr[].

   however vranlibm must accomodate both type 5 and 6 files;
   type 6 has extra stuff after the seq_id.
*/

void
vranlibm(str,cnt,libpos)
	char *str; int cnt;
	long libpos;
{
  char *bp, *mp;
  int llen;

  llen = m_file.s_pos_arr[libpos]-m_file.d_pos_arr[libpos];

  mp = m_file.mmap_base+m_file.d_pos_arr[libpos];
  
  strncpy(str,mp+4,20);
  str[20]='\0';
  if ((bp=strchr(str,' '))!=NULL) *(bp+1) = '\0';
  else if ((bp=strchr(str,'\n'))!=NULL) *bp = ' ';
  bp = strchr(mp,'\n');

  llen -= (bp-mp)-5;
  if (llen >  cnt-strlen(str)) llen = cnt-strlen(str)-1;

  strncat(str,bp+1,llen);
  if ((bp = strchr(str,'\n'))!=NULL) *bp='\0';
  str[cnt-1]='\0';
  lpos = libpos;
}

void
close_mmap() {
  free(m_file.s_pos_arr);
  free(m_file.d_pos_arr);
  if (m_file.mm_flg) {
    munmap(m_file.mmap_base,m_file.st_size);
    free(m_fptr);
  }
}  

#ifndef min
#define min(x,y) ((x) > (y) ? (y) : (x))
#endif

static char gcg_type[10];
static long gcg_len;
static int gcg_bton[4]={2,4,1,3};

gcg_getlibm(seq,maxs,libstr,libpos,lcont)
     unsigned char *seq;
     int maxs;
     char *libstr;
     long *libpos;
     int *lcont;
{
  char dummy[20];
  char gcg_date[6];
  register unsigned char *cp, *seqp, stmp;
  register int *ap;
  unsigned char *seqm, *seqm1;
  long r_block, b_block, r_fact, r16_block;

  seqp = seq;
  seqm = &seq[maxs-9];
  seqm1 = seqm-1;
#if !defined(TFAST)
  ap = sascii;
#else
  ap = nascii;
#endif
  if (*lcont==0) {
    if (lpos >= m_file.max_cnt) return (-1);
    sscanf(m_file.mmap_base+m_file.d_pos_arr[lpos]+4,"%s %s %s %s %ld\n",
	   libstr,gcg_date,gcg_type,dummy,&gcg_len);

    libstr[12]='\0';
    *libpos = lpos;
    m_file.mmap_addr = m_file.mmap_base+m_file.s_pos_arr[lpos];
  }

  r_block = b_block = min((size_t)(seqm-seqp),gcg_len);
  if (gcg_type[0]=='2') {
    r_block = (r_block+3)/4;
  }

  cp=(unsigned char *)m_file.mmap_addr; 
  if (gcg_type[0]=='A') {
    r_fact = 1;
    r16_block = r_block/16;
    while (r16_block-- > 0) {
      *seqp++ = ap[*cp++];
      *seqp++ = ap[*cp++];
      *seqp++ = ap[*cp++];
      *seqp++ = ap[*cp++];
      *seqp++ = ap[*cp++];
      *seqp++ = ap[*cp++];
      *seqp++ = ap[*cp++];
      *seqp++ = ap[*cp++];
      *seqp++ = ap[*cp++];
      *seqp++ = ap[*cp++];
      *seqp++ = ap[*cp++];
      *seqp++ = ap[*cp++];
      *seqp++ = ap[*cp++];
      *seqp++ = ap[*cp++];
      *seqp++ = ap[*cp++];
      *seqp++ = ap[*cp++];
    }
    while (seqp<seq+r_block) *seqp++ = ap[*cp++];
  }
  else if (gcg_type[0]=='2') {
    r_fact = 4;
    r16_block = r_block/8;
    while(r16_block-- > 0) {
      stmp = *cp++;
      *seqp++ = gcg_bton[(stmp>>6) &3];
      *seqp++ = gcg_bton[(stmp>>4) &3];
      *seqp++ = gcg_bton[(stmp>>2) &3];
      *seqp++ = gcg_bton[(stmp) &3];
      stmp = *cp++;
      *seqp++ = gcg_bton[(stmp>>6) &3];
      *seqp++ = gcg_bton[(stmp>>4) &3];
      *seqp++ = gcg_bton[(stmp>>2) &3];
      *seqp++ = gcg_bton[(stmp) &3];
      stmp = *cp++;
      *seqp++ = gcg_bton[(stmp>>6) &3];
      *seqp++ = gcg_bton[(stmp>>4) &3];
      *seqp++ = gcg_bton[(stmp>>2) &3];
      *seqp++ = gcg_bton[(stmp) &3];
      stmp = *cp++;
      *seqp++ = gcg_bton[(stmp>>6) &3];
      *seqp++ = gcg_bton[(stmp>>4) &3];
      *seqp++ = gcg_bton[(stmp>>2) &3];
      *seqp++ = gcg_bton[(stmp) &3];
      stmp = *cp++;
      *seqp++ = gcg_bton[(stmp>>6) &3];
      *seqp++ = gcg_bton[(stmp>>4) &3];
      *seqp++ = gcg_bton[(stmp>>2) &3];
      *seqp++ = gcg_bton[(stmp) &3];
      stmp = *cp++;
      *seqp++ = gcg_bton[(stmp>>6) &3];
      *seqp++ = gcg_bton[(stmp>>4) &3];
      *seqp++ = gcg_bton[(stmp>>2) &3];
      *seqp++ = gcg_bton[(stmp) &3];
      stmp = *cp++;
      *seqp++ = gcg_bton[(stmp>>6) &3];
      *seqp++ = gcg_bton[(stmp>>4) &3];
      *seqp++ = gcg_bton[(stmp>>2) &3];
      *seqp++ = gcg_bton[(stmp) &3];
      stmp = *cp++;
      *seqp++ = gcg_bton[(stmp>>6) &3];
      *seqp++ = gcg_bton[(stmp>>4) &3];
      *seqp++ = gcg_bton[(stmp>>2) &3];
      *seqp++ = gcg_bton[(stmp) &3];
    }

    while (seqp < seq+4*r_block) {
      stmp = *cp++;
      *seqp++ = gcg_bton[(stmp>>6) &3];
      *seqp++ = gcg_bton[(stmp>>4) &3];
      *seqp++ = gcg_bton[(stmp>>2) &3];
      *seqp++ = gcg_bton[(stmp) &3];
    }
  }
  if (r_fact * r_block >= gcg_len) {
    *lcont = 0;
    lpos++;
  }
  else {
    if (gcg_type[0]=='2') b_block = 4*r_block;
    gcg_len -= b_block;
    (*lcont)++;
  }

  seq[b_block] = EOSEQ;
  /*   if (b_block==0) return 1; else */
  return b_block;
}

lgetlibm(seq,maxs,libstr,libpos,lcont)
     unsigned char *seq;
     int maxs;
     char *libstr;
     long *libpos;
     int *lcont;
{
  register unsigned char *cp, *seqp;
  register int *ap;
  unsigned char *seqm, *seqm1;

  seqp = seq;
  seqm = &seq[maxs-11];
  seqm1 = seqm-1;
#if !defined(TFAST)
  ap = sascii;
#else
  ap = nascii;
#endif
  if (*lcont==0) {
    if (lpos >= m_file.max_cnt) return (-1);

    strncpy(libstr,m_file.mmap_base+m_file.d_pos_arr[lpos]+12,12);
    libstr[12]='\0';
    *libpos = lpos;

    m_file.mmap_addr = m_file.mmap_base+m_file.s_pos_arr[lpos];
    cp = (unsigned char *)m_file.mmap_addr;
  }
  else cp = (unsigned char *)m_file.mmap_addr;

  while (seqp<seqm1) {
    if (*cp=='/' && *(cp-1)=='\n') break;
    if ((*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA) continue;
    --seqp;
    if (*cp=='\n' && *(cp+1)==' ') cp += 11;
  }

  if (seqp>=seqm1) {
    (*lcont)++;
    m_file.mmap_addr = (char *)cp;
  }
  else {
    *lcont=0;
    lpos++;
  }

  *seqp = EOSEQ;
  return (int)(seqp-seq);
}

/* this code seeks to provide both the various accession numbers
   necessary to identify the sequence, and also some description.

   Unfortunately, the various contributors to Genbank use three
   slightly different formats for including the accession number.

(1)LOCUS       HSJ214M20  107422 bp    DNA             HTG       16-JUN-2000
   DEFINITION  Homo sapiens chromosome 6 clone RP1-214M20 map p12.1-12.3, ***
               SEQUENCING IN PROGRESS ***, in unordered pieces.
   ACCESSION   AL121969

(2)LOCUS       AL359201   117444 bp    DNA             HTG       15-JUN-2000
   DEFINITION  Homo sapiens chromosome 1 clone RP4-671C13 map p13.2-21.1, ***
               SEQUENCING IN PROGRESS ***, in unordered pieces.
   ACCESSION   AL359201

(3)LOCUS       BB067000      280 bp    mRNA            EST       19-JUN-2000
   DEFINITION  BB067000 RIKEN full-length enriched, 15 days embryo male testis Mus
               musculus cDNA clone 8030456L01 3', mRNA sequence.
   ACCESSION   BB067000

This makes it more difficult to both provide the accession number in a
standard location and to conserve definition space
*/

void
lranlibm(str,cnt,seek)
     char *str; int cnt; long seek;
{
  char *bp, *llp;
  char lline[MAXLINE], acc[MAXLINE], desc[MAXLINE];
  int ll;

  llp = m_file.mmap_addr = m_file.mmap_base + m_file.d_pos_arr[seek];

  strncpy(str,llp+12,12);
  if ((bp=strchr(str,' '))!=NULL) *(bp+1) = '\0';
  str[12]='\0';
  mgets(desc,MAXLINE);

  while (desc[0]!='D' || desc[1]!='E' || strncmp(desc,"DEFINITION",10))
    mgets(desc,MAXLINE);

  if ((bp = strchr(&desc[12],'\n'))!=NULL) *bp='\0';

  /* get the accession line */
  strncpy(acc,desc,sizeof(acc));
  while (acc[0]!='A' || acc[1]!='C' || strncmp(acc,"ACCESSION",9))
    mgets(acc,MAXLINE);
  if ((bp = strchr(&acc[12],'\n'))!=NULL) *bp='\0';
  if ((bp = strchr(&acc[12],' '))!=NULL) *bp='\0';

  /* add accession to locus */
  if ((bp=strchr(&desc[12],' '))!=NULL) {
    *bp = '\0';
    if (strcmp(&acc[12],&desc[12])==0) {
      *bp = ' ';
      strncat(str,&desc[11],cnt-strlen(str));
    }
    else {
      *bp = ' ';
      strncat(str,&acc[11],cnt-strlen(str));
      strncat(str,&desc[11],cnt-strlen(str));
    }
  }
  else {
    strncat(str,&acc[11],cnt-strlen(str));
    strncat(str,&desc[11],cnt-strlen(str));
  }    
  str[cnt-1]='\0';

  lpos = seek;
}

