/*	ncbl2_lib.c	functions to read ncbi-blast format files from
			formatdb (blast2.0 format files)

		copyright (c) 1999 William R. Pearson
*/

/* $Name: fa33t09d0 $ - $Id: ncbl2_mlib.c,v 1.8.2.2.2.1 2001/04/05 14:16:08 wrp Exp $ */

/* to turn on mmap()ing for Blast2 files: */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

/* ****************************************************************
This code reads NCBI Blast2 format databases from formatdb version 3

(From NCBI) This section describes the format of the databases.

Formatdb creates three main files for proteins containing indices,
sequences, and headers with the extensions, respectively, of pin, psq,
and phr (for nucleotides these are nin, nsq, and nhr).  A number of
other ISAM indices are created, but these are described elsewhere.

FORMAT OF THE INDEX FILE
------------------------

1.) formatdb version number 	[4 bytes].

2.) protein dump flag (1 for a protein database, 0 for a nucleotide
    database) [4 bytes].

3.) length of the database title in bytes	[4 bytes].
4.) the database title		[length given in 3.)].
5.) length of the date/time string	[4 bytes].
6.) the date/time string	[length given in 5.)].
7.) the number of sequences in the database	[4 bytes].
8.) the total length of the database in residues/basepairs	[4 bytes].
9.) the length of the longest sequence in the database 		[4 bytes].

10.) a list of the offsets for definitions (one for each sequence) in
the header file.  There are num_of_seq+1 of these, where num_of_seq is
the number of sequences given in 7.).

11.) a list of the offsets for sequences (one for each sequence) in
the sequence file.  There are num_of_seq+1 of these, where num_of_seq
is the number of sequences given in 7.).

12.) a list of the offsets for the ambiguity characters (one for each
sequence) in the sequence file.  This list is only present for
nucleotide databases and, since the database is compressed 4/1 for
nucleotides, allows the ambiguity characters to be restored when the
sequence is generated.  There are num_of_seq+1 of these, where
num_of_seq is the number of sequences given in 7.).


FORMAT OF THE SEQUENCE FILE
---------------------------

There are different formats for the protein and nucleotide sequence files.

The protein sequence files is quite simple.  The first byte in the
file is a NULL byte, followed by the sequence in ncbistdaa format
(described in the NCBI Software Development Toolkit documentation).
Following the sequence is another NULL byte, followed by the next
sequence.  The file ends with a NULL byte, following the last
sequence.

The nucleotide sequence file contains the nucleotide sequence, with
four basepairs compressed into one byte.  The format used is NCBI2na,
documented in the NCBI Software Development Toolkit manual.  Any
ambiguity characters present in the original sequence are replaced at
random by A, C, G or T.  The true value of ambiguity characters are
stored at the end of each sequence to allow true reproduction of the
original sequence.

FORMAT OF THE HEADER FILE
-------------------------

The format of the header file depends on whether or not the identifiers in the
original file were parsed or not.  For the case that they were not, then each
entry has the format:

gnl|BL_ORD_ID|entry_number my favorite yeast sequence...

Here entry_number gives the ordinal number of the sequence in the
database (with zero offset).  The identifier
gnl|BL_ORD_ID|entry_number is used by the BLAST software to identify
the entry, if the user has not provided another identifier.  If the
identifier was parsed, then gnl|BL_ORD_ID|entry_number is replaced by
the correct identifier, as described in
ftp://ncbi.nlm.nih.gov/blast/db/README .

There are no separators between these deflines.

**************************************************************** */


#ifdef UNIX
#define RBSTR "r"
#else
#define RBSTR "rb"
#endif

#define XTERNAL
#include "uascii.h"
#include "upam.h"
#include "ncbl2_head.h"

unsigned int bl2_uint4_cvt(unsigned int);
unsigned int bl2_long4_cvt(long);
long bl2_long8_cvt(long);
void src_int4_read(FILE *fd, int * );
void src_uint4_read(FILE *fd, unsigned int * );
void src_long4_read(FILE *fd, long *);
void src_ulong4_read(FILE *fd, unsigned long *);
void src_char_read(FILE *fd, char *);

/* nt_btoa maps  from blast 2bit  format to ascii  characters */
static char nt_btoa[5] = {"ACGT"};

static char aa_b2toa[27]= {"-ABCDEFGHIKLMNPQRSTVWXYZU*"};

static int aa_btof[32];	/* maps to fasta alphabet */

static FILE *ifile=NULL,	/* index offsets, also DB info */
	    *hfile=NULL,	/* description lines */
	    *sfile=NULL;	/* binary sequence data */

static unsigned int lib_cnt;
static int dbtype, dbformat, amb_cnt;
static char dline[512];
static long bl_lib_pos;

#define NCBIBL20 12

extern int (*getlib)();
int ncbl2_mgetliba(char *, int, char *, long *, int *);
int ncbl2_mgetlibn(char *, int, char *, long *, int *);

int ncbl2_getliba(char *, int, char *, long *, int *);
int ncbl2_getlibn(char *, int, char *, long *, int *);
void newname(char *, char *, char *, int);

#ifdef USE_MMAP
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#ifdef IBM_AIX
#include <fcntl.h>
#else
#include <sys/fcntl.h>
#endif
#endif


#ifdef USE_MMAP
#ifndef MAP_FILE
#define MAP_FILE 0
#endif
#endif
#include "mm_file.h"
static struct m_file_str *m_fptr, m_file;
 
void ncbl2_ranlib(char *, int, long);

/* ncbl2_openlib() is used to open (and memory map) a BLAST2.0 format
   file.  Ifdef USE_MMAP, then ncbl2_openlib returns a structure that can
   be used to read the database.  If we are not using mmap(), then
   ncbl2_openlib returns 0 for failure and 1 for successs */
   
#ifdef USE_MMAP
struct m_file_str *
#else
int
#endif
ncbl2_openlib(char *name, int ldnaseq)
{
  char hname[256];
  char sname[256];
  char tname[256];
  int title_len;
  char *title_str=NULL;
  int date_len;
  char *date_str=NULL;
  int tmp;
  int i;
  unsigned long line_len, c_len, clean_count;
#ifdef USE_MMAP
  struct stat statbuf;
#endif
  unsigned int *f_pos_arr;

  /* make certain that m_file is all 0's whenever called */
  memset(&m_file,0,sizeof(m_file));

  /* ncbl2_ranlib is used for all BLAST2.0 access */
  m_file.ranlib = ncbl2_ranlib;

  if (!ldnaseq) {	/* read a protein database */
    newname(tname,name,AA_INDEX_EXT,(int)sizeof(tname));
    newname(hname,name,AA_HEADER_EXT,(int)sizeof(hname));
    newname(sname,name,AA_SEARCHSEQ_EXT,(int)sizeof(sname));

    /* getlib access for proteins */
    m_file.getlib = getlib = ncbl2_getliba;

    /* initialize map of BLAST2 amino acids to FASTA amino acids */
    for (i=0; i<sizeof(aa_b2toa); i++) {
      if ((tmp=aascii[aa_b2toa[i]])<NA) aa_btof[i]=tmp;
      else if (aa_b2toa[i]=='*') aa_btof[i]=aascii['X'];
      else aa_b2toa[i]=0;
/*    else aa_btof[i]=aascii['X']; */
    }
  }
  else {	/* reading DNA library */
    newname(tname,name,NT_INDEX_EXT,(int)sizeof(tname));
    newname(hname,name,NT_HEADER_EXT,(int)sizeof(hname));
    newname(sname,name,NT_SEARCHSEQ_EXT,(int)sizeof(sname));

    m_file.getlib = getlib = ncbl2_getlibn;
  }
	
  /* open the index file */
  if ((ifile = fopen(tname,RBSTR))==NULL) {
    fprintf(stderr," cannot open %s (%s) INDEX file",tname,name);
    perror("...");
    return 0;
  }
  src_int4_read(ifile,&dbformat);	/* get format DB version number */
  src_int4_read(ifile,&dbtype); 	/* get 1 for protein/0 DNA */

  if (dbformat != FORMATDBV) {
    fprintf(stderr,"error - %s wrong formatdb version (%d/%d)\n",
	    tname,dbformat,FORMATDBV);
    return 0;
  }

  if ((!ldnaseq && dbtype != AAFORMAT) || (ldnaseq && dbtype!=NTFORMAT)) {
    fprintf(stderr,"error - %s wrong format (%d/%d)\n",
	    tname,dbtype,(ldnaseq ? NTFORMAT: AAFORMAT));
    return 0;
  }

  /* open the header file */
  if ((hfile = fopen(hname,RBSTR))==NULL) {
    fprintf(stderr," cannot open %s header file\n",hname);
    return 0;
  }

  /* open the sequence file */

#if defined (USE_MMAP) 
  m_file.mm_flg=((m_file.mmap_fd=open(sname,O_RDONLY))>=0);
  if (!m_file.mm_flg) {
    fprintf(stderr," cannot open %s",sname);
    perror("...");
  }
  else {
    if(fstat(m_file.mmap_fd, &statbuf) < 0) {
      fprintf(stderr," cannot fstat %s",sname);
      perror("...");
      m_file.mm_flg = 0;
    }
    else {
      m_file.st_size = statbuf.st_size;
      if((m_file.mmap_base = 
	  mmap(NULL, m_file.st_size, PROT_READ,
	       MAP_FILE | MAP_SHARED, m_file.mmap_fd, 0)) == (char *) -1) {
	fprintf(stderr," cannot mmap %s",sname);
	perror("...");
	m_file.mm_flg = 0;
      }  
      else {
	m_file.mmap_addr = m_file.mmap_base;
	m_file.mm_flg = 1;
      }
    }
    /* regardless, close the open()ed version */
    close(m_file.mmap_fd);
  }
#else
  m_file.mm_flg = 0;
#endif

  if  (!m_file.mm_flg) {
    if ((sfile = fopen(sname,RBSTR))==NULL) {
      fprintf(stderr," cannot open %s sequence file",sname);
      perror("...");
      return 0;
    }
  }

/* all files should be open */

  src_int4_read(ifile,&title_len);
  if (title_len > 0) {
    if ((title_str = calloc((size_t)title_len+1,sizeof(char)))==NULL) {
      fprintf(stderr," cannot allocate title string (%d)\n",title_len);
      return 0;
    }
    fread(title_str,(size_t)1,(size_t)title_len,ifile);
  }
  
  src_int4_read(ifile,&date_len);
  if (date_len > 0) {
    if ((date_str = calloc((size_t)date_len+1,sizeof(char)))==NULL) {
      fprintf(stderr," cannot allocate date string (%d)\n",date_len);
      return 0;
    }
    fread(date_str,(size_t)1,(size_t)date_len,ifile);
  }
  
  lib_cnt = 0;
  src_int4_read(ifile,&m_file.max_cnt);
  src_ulong4_read(ifile,&m_file.tot_len);
  src_ulong4_read(ifile,&m_file.max_len);

  /* currently we are not using this information, but perhaps later */
  if (title_str!=NULL) free(title_str);
  if (date_str!=NULL) free(date_str);

#ifdef DEBUG
    fprintf(stderr,"%s format: BL2 (%s)  max_cnt: %d, totlen: %ld, maxlen %ld\n",
	    name,m_file.mm_flg ? "mmap" : "fopen", 
	    m_file.max_cnt,m_file.tot_len,m_file.max_len);
#endif

  /* allocate and read hdr indexes */
  if ((f_pos_arr=(unsigned int *)calloc((size_t)m_file.max_cnt+1,sizeof(int)))==NULL) {
      fprintf(stderr," cannot allocate tmp header pointers\n");
      return 0;
    }

  if ((m_file.d_pos_arr=(unsigned long *)calloc((size_t)m_file.max_cnt+1,sizeof(long)))==NULL) {
      fprintf(stderr," cannot allocate header pointers\n");
      return 0;
    }

  /* allocate and read sequence offsets */
  if ((m_file.s_pos_arr=(unsigned long *)calloc((size_t)m_file.max_cnt+1,sizeof(long)))==NULL) {
      fprintf(stderr," cannot allocate sequence pointers\n");
      return 0;
    }

  /*
  for (i=0; i<=m_file.max_cnt; i++) src_uint4_read(ifile,&m_file.d_pos_arr[i]);
  for (i=0; i<=m_file.max_cnt; i++) src_uint4_read(ifile,&m_file.s_pos_arr[i]);
  */
  if (fread(f_pos_arr,(size_t)4,m_file.max_cnt+1,ifile)!=m_file.max_cnt+1) {
    fprintf(stderr," error reading hdr offsets: %s\n",tname);
    return 0;
  }

  for (i=0; i<=m_file.max_cnt; i++)
#ifdef IS_BIG_ENDIAN
    m_file.d_pos_arr[i] = f_pos_arr[i];
#else
    m_file.d_pos_arr[i] = bl2_uint4_cvt(f_pos_arr[i]);
#endif

  if (fread(f_pos_arr,(size_t)4,m_file.max_cnt+1,ifile)!=m_file.max_cnt+1) {
    fprintf(stderr," error reading seq offsets: %s\n",tname);
    return 0;
  }
  for (i=0; i<=m_file.max_cnt; i++) {
#ifdef IS_BIG_ENDIAN
    m_file.s_pos_arr[i] = f_pos_arr[i];
#else
    m_file.s_pos_arr[i] = bl2_uint4_cvt(f_pos_arr[i]);
#endif
  }
  if (dbtype == NTFORMAT) {
    /* allocate and ambiguity  offsets */
    if ((m_file.a_pos_arr=(unsigned long *)calloc((size_t)m_file.max_cnt+1,sizeof(long)))==NULL) {
      fprintf(stderr," cannot allocate sequence pointers\n");
      return 0;
    }

    /*
    for (i=0; i<=m_file.max_cnt; i++) src_uint4_read(ifile,&m_file.a_pos_arr[i]);
    */

    if (fread(f_pos_arr,(size_t)4,m_file.max_cnt+1,ifile)!=m_file.max_cnt+1) {
      fprintf(stderr," error reading seq offsets: %s\n",tname);
      return 0;
    }
    for (i=0; i<=m_file.max_cnt; i++) {
#ifdef IS_BIG_ENDIAN
      m_file.a_pos_arr[i] = f_pos_arr[i];
#else
      m_file.a_pos_arr[i] = bl2_uint4_cvt(f_pos_arr[i]);
#endif
    }
  }

  free(f_pos_arr);

  if (!m_file.mm_flg) {
    tmp = fgetc(sfile);
    if (tmp!=NULLB)
      fprintf(stderr," phase error: %d:%d found\n",lib_cnt,tmp);
  }

  bl_lib_pos = 1;
  amb_cnt = 0;

#ifdef USE_MMAP  
  if ((m_fptr=(struct m_file_str *)calloc(1,sizeof(struct m_file_str)))==NULL) {
    fprintf(stderr," cannot allocate m_file_str\n");
    return 0;
  }
  else {
    memcpy(m_fptr,&m_file,sizeof(struct m_file_str));
    return m_fptr;
  }
#else
  return 1;
#endif
}

void ncbl2_closelib()
{
  if (ifile !=NULL ) {
    fclose(ifile); ifile=NULL;
    free(m_file.s_pos_arr); m_file.s_pos_arr = NULL;
    if (m_file.a_pos_arr!=NULL) {
      free(m_file.a_pos_arr);
      m_file.a_pos_arr = NULL;
    }
  }
  if (hfile !=NULL ) {
    fclose(hfile); hfile=NULL;
    free(m_file.d_pos_arr); m_file.d_pos_arr = NULL;
  }

#ifdef USE_MMAP
  if (m_file.mm_flg) {
    munmap(m_file.mmap_base,m_file.st_size);
    m_file.mmap_fd = -1;
  }
  else 
#endif
    if (sfile !=NULL ) {fclose(sfile); sfile=NULL;}
}

void
reopen_ncbl2(struct m_file_str *m_fptr) {
#ifdef USE_MMAP
  memcpy(&m_file,m_fptr,sizeof(struct m_file_str));
#else
  m_file.mm_flg = 0;
#endif
  lib_cnt = 0;
}

#define min(a,b) ((a) < (b) ) ? (a) : (b)

int
ncbl2_getliba(seq,maxs,libstr,libpos,lcont)
     char *seq;
     int maxs;
     char *libstr;
     long *libpos;
     int *lcont;
{
  register char *sptr, *dptr;
  int s_chunk;
  int desc_len;
  long seqcnt;
  long tmp;
  char ch, *bp;
  static long seq_len;
  
  *libpos = lib_cnt;
  if (*lcont==0) {
    if (lib_cnt >= m_file.max_cnt) return -1;	/* no more sequences */
    seq_len = m_file.s_pos_arr[lib_cnt+1] - m_file.s_pos_arr[lib_cnt]; /* value is +1 off to get the NULL */
    if (m_file.mm_flg) m_file.mmap_addr = m_file.mmap_base+m_file.s_pos_arr[lib_cnt];
#if !defined(DEBUG) && !defined(PCOMPLIB)
    libstr[0]='\0';
#else
    /* get the name from the header file */
    fseek(hfile,m_file.d_pos_arr[lib_cnt],0);
#ifndef PCOMPLIB
    fread(libstr,(size_t)1,(size_t)12,hfile);
    libstr[12]='\0';
    if ((bp=strchr(libstr,' '))!=NULL) *bp='\0';
#else
    desc_len = min(m_file.d_pos_arr[lib_cnt+1]-m_file.d_pos_arr[lib_cnt],255);
    fread(libstr,(size_t)1,(size_t)desc_len,hfile);
    libstr[255]='\0';
    while ((bp = strchr(libstr,'\001'))!=NULL ||
	   (bp = strchr(libstr,'\t'))!=NULL) *bp=' ';
#endif      
#endif
    }
  

  if (seq_len <= maxs) { /* sequence fits */
    seqcnt = seq_len;
    lib_cnt++;
    *lcont = 0;
  }
  else {		/* doesn't fit */
    seqcnt = maxs-1;
    (*lcont)++;
  } 

  if (m_file.mm_flg) sptr = m_file.mmap_addr;
  else {
    if ((tmp=fread(seq,(size_t)1,(size_t)seq_len,sfile))!=(size_t)seq_len) {
      fprintf(stderr," could not read sequence record: %ld %ld != %ld\n",
	      *libpos,tmp,seq_len);
      goto error; 
    }
    sptr = seq;
  }
  if (seq_len <= maxs) {seqcnt = --seq_len;}

  /* everything is ready, set up dst. pointer, seq_len */
  dptr = seq;

  if (aa_b2toa[sptr[seq_len-1]]=='*') seq_len--;
  s_chunk = seqcnt/16;
  while (s_chunk-- > 0) {
    *dptr++ = aa_btof[*sptr++];
    *dptr++ = aa_btof[*sptr++];
    *dptr++ = aa_btof[*sptr++];
    *dptr++ = aa_btof[*sptr++];
    *dptr++ = aa_btof[*sptr++];
    *dptr++ = aa_btof[*sptr++];
    *dptr++ = aa_btof[*sptr++];
    *dptr++ = aa_btof[*sptr++];
    *dptr++ = aa_btof[*sptr++];
    *dptr++ = aa_btof[*sptr++];
    *dptr++ = aa_btof[*sptr++];
    *dptr++ = aa_btof[*sptr++];
    *dptr++ = aa_btof[*sptr++];
    *dptr++ = aa_btof[*sptr++];
    *dptr++ = aa_btof[*sptr++];
    *dptr++ = aa_btof[*sptr++];
  }
  while (dptr < seq+seqcnt) *dptr++ = aa_btof[*sptr++];

  if (m_file.mm_flg) m_file.mmap_addr = sptr;

  /* we didn't get it all, so reset for more */
  if (*lcont) seq_len -= seqcnt;

  seq[seqcnt]= EOSEQ;
  return (seqcnt);
  
error:	fprintf(stderr," error reading %ld at %ld\n",libstr,*libpos);
  fflush(stderr);
  return (-1);
}

char tmp_amb[4096];

ncbl2_getlibn(seq,maxs,libstr,libpos,lcont)
     char *seq;
     int maxs;
     char *libstr;
     long *libpos;
     int *lcont;
{
  register char *sptr, *tptr, stmp;
  long seqcnt;
  int s_chunk, desc_len;
  size_t tmp;
  char ch;
  static long seq_len;
  static int c_len,c_pad;
  int c_len_set;
  char *bp;

  *libpos = lib_cnt;
  if (*lcont==0) {	/* not a continuation of previous */
    if (lib_cnt >= m_file.max_cnt) return (-1);
    c_len = m_file.a_pos_arr[lib_cnt]- m_file.s_pos_arr[lib_cnt];
    if (!m_file.mm_flg) {
      if (bl_lib_pos != m_file.s_pos_arr[lib_cnt]) { /* are we positioned to read? */
	amb_cnt++;
	if ((bl_lib_pos - m_file.s_pos_arr[lib_cnt]) < sizeof(tmp_amb)) {
	  /* jump over amb_ray */
	  fread(tmp_amb,(size_t)1,(size_t)(m_file.s_pos_arr[lib_cnt]-bl_lib_pos),sfile);
	}
	else {	/* fseek over amb_ray */
	  fseek(sfile,m_file.s_pos_arr[lib_cnt],0);
	}
	bl_lib_pos = m_file.s_pos_arr[lib_cnt];
      }
    }
    else m_file.mmap_addr = m_file.mmap_base + m_file.s_pos_arr[lib_cnt];
#if !defined(DEBUG) && !defined(PCOMPLIB)
    libstr[0]='\0';
#else
    /* get the name from the header file */
    fseek(hfile,m_file.d_pos_arr[lib_cnt],0);
#ifndef PCOMPLIB
    fread(libstr,(size_t)1,(size_t)12,hfile);
    libstr[12]='\0';
    if ((bp=strchr(libstr,' '))!=NULL) *bp='\0';
#else
    desc_len = min(m_file.d_pos_arr[lib_cnt+1]-m_file.d_pos_arr[lib_cnt],255);
    fread(libstr,(size_t)1,(size_t)desc_len,hfile);
    libstr[255]='\0';
    while ((bp = strchr(libstr,'\001'))!=NULL ||
	   (bp = strchr(libstr,'\t'))!=NULL) *bp=' ';
#endif      
#endif      
  }			/* end of *lcont==0 */

  /* To avoid the situation where c_len <= 1; we must anticipate what
     c_len will be after this pass.  If it will be <= 64, back off this
     time so next time it will be > 64 */

  seq_len = c_len*4;

  if ((seq_len+4 > maxs) && (seq_len+4 - maxs  <= 256)) {
    /* we won't be done but we will have less than 256 to go */
    c_len -= 64; seq_len -= 256; c_len_set = 1; maxs -= 256;}
  else c_len_set = 0;

  /*
  fprintf(stderr," lib_cnt: %d %d %d %d\n",lib_cnt,c_len,seq_len,maxs);
  */

  /* does the rest of the sequence fit? */
  if (seq_len <= maxs-4 && !c_len_set) {
    seqcnt = c_len;
    if (!m_file.mm_flg) {
      if ((tmp=fread(seq,(size_t)1,(size_t)seqcnt,sfile))!=(size_t)seqcnt) {
	fprintf(stderr,
		" could not read sequence record: %s %ld %ld != %ld: %d\n",
		libstr,*libpos,tmp,seqcnt,*seq);
	goto error; 
      }
      bl_lib_pos += tmp;
      sptr = seq + seqcnt;
    }
    else sptr = m_file.mmap_addr+seqcnt;

    *lcont = 0;		/* this is the last chunk */
    lib_cnt++;		/* increment to the next sequence */
    /* the last byte is either '0' (no remainder) or the last 1-3 chars and the remainder */
    c_pad = *(sptr-1);
    c_pad &= 0x3;	/* get the last (low) 2 bits */
    seq_len -= (4 - c_pad);	/* if the last 2 bits are 0, its a NULL byte */
  }
  else {	/* get the next chunk, but more to come */
    seqcnt = ((maxs+3)/4)-1;
    if (!m_file.mm_flg) {
      if ((tmp=fread(seq,(size_t)1,(size_t)(seqcnt),sfile))!=(size_t)(seqcnt)) {
	fprintf(stderr," could not read sequence record: %ld %ld/%ld\n",
		*libpos,tmp,seqcnt);
	goto error;
      }
      bl_lib_pos += tmp;
      sptr = seq + seqcnt;
    }
    else {
      sptr = m_file.mmap_addr+seqcnt;
      m_file.mmap_addr += seqcnt;
    }
    seq_len = 4*seqcnt;
    c_len -= seqcnt;
    if (c_len_set) {c_len += 64; maxs += 256;}
    (*lcont)++;
/*  hopefully we don't need this because of c_len -= 64. */
/*
    if (c_len == 1) {
#if !defined (USE_MMAP)
      c_pad = fgetc(sfile);
      *sptr=c_pad;
#else
      c_pad = *m_file.mmap_addr++;
      sptr = m_file.mmap_addr;
#endif
      c_pad &= 0x3;
      seq_len += c_pad;
      seqcnt++;
      lib_cnt++;
      *lcont = 0;
    }
*/
  }

  /* point to the last packed byte and to the end of the array
     seqcnt is the exact number of bytes read
     tptr points to the destination, use multiple of 4 to simplify math
     sptr points to the source, note that the last byte will be read 4 cycles
     before it is written
     */
  
  tptr = seq + 4*seqcnt;
  s_chunk = seqcnt/8;
  while (s_chunk-- > 0) {
    stmp = *--sptr;
    *--tptr = (stmp&3) +1;
    *--tptr = ((stmp >>= 2)&3)+1;
    *--tptr = ((stmp >>= 2)&3)+1;
    *--tptr = ((stmp >>= 2)&3)+1;
    stmp = *--sptr;
    *--tptr = (stmp&3) +1;
    *--tptr = ((stmp >>= 2)&3)+1;
    *--tptr = ((stmp >>= 2)&3)+1;
    *--tptr = ((stmp >>= 2)&3)+1;
    stmp = *--sptr;
    *--tptr = (stmp&3) +1;
    *--tptr = ((stmp >>= 2)&3)+1;
    *--tptr = ((stmp >>= 2)&3)+1;
    *--tptr = ((stmp >>= 2)&3)+1;
    stmp = *--sptr;
    *--tptr = (stmp&3) +1;
    *--tptr = ((stmp >>= 2)&3)+1;
    *--tptr = ((stmp >>= 2)&3)+1;
    *--tptr = ((stmp >>= 2)&3)+1;
    stmp = *--sptr;
    *--tptr = (stmp&3) +1;
    *--tptr = ((stmp >>= 2)&3)+1;
    *--tptr = ((stmp >>= 2)&3)+1;
    *--tptr = ((stmp >>= 2)&3)+1;
    stmp = *--sptr;
    *--tptr = (stmp&3) +1;
    *--tptr = ((stmp >>= 2)&3)+1;
    *--tptr = ((stmp >>= 2)&3)+1;
    *--tptr = ((stmp >>= 2)&3)+1;
    stmp = *--sptr;
    *--tptr = (stmp&3) +1;
    *--tptr = ((stmp >>= 2)&3)+1;
    *--tptr = ((stmp >>= 2)&3)+1;
    *--tptr = ((stmp >>= 2)&3)+1;
    stmp = *--sptr;
    *--tptr = (stmp&3) +1;
    *--tptr = ((stmp >>= 2)&3)+1;
    *--tptr = ((stmp >>= 2)&3)+1;
    *--tptr = ((stmp >>= 2)&3)+1;
  }
  while (tptr>seq) {
    stmp = *--sptr;
    *--tptr = (stmp&3) +1;
    *--tptr = ((stmp >>= 2)&3)+1;
    *--tptr = ((stmp >>= 2)&3)+1;
    *--tptr = ((stmp >>= 2)&3)+1;
  }
  /*
    for (sptr=seq; sptr < seq+seq_len; sptr++) {
    printf("%c",nt[*sptr]);
    if ((int)(sptr-seq) % 60 == 59) printf("\n");
    }
    printf("\n");
    */


  if (seqcnt*4 >= seq_len) {	/* there was enough room */
    seq[seq_len]= EOSEQ;
    /* printf("%d\n",seq_len); */
    return seq_len;
  }
  else {				/* not enough room */
    seq[seqcnt*4]=EOSEQ;
    seq_len -= 4*seqcnt;
    return (4*seqcnt);
  }
  
error:	fprintf(stderr," error reading %ld at %ld\n",libstr,*libpos);
  fflush(stderr);
  return (-1);
}

void
ncbl2_ranlib(str,cnt,libpos)
	char *str; int cnt;
	long libpos;
{
  int llen;
  char *bp;

  lib_cnt = libpos;
  llen = m_file.d_pos_arr[libpos+1]-m_file.d_pos_arr[libpos];
  if (llen >= cnt) llen = cnt-1;
  fseek(hfile,m_file.d_pos_arr[libpos],0);

  fread(str,(size_t)1,(size_t)(llen),hfile);
  str[llen]='\0';

  bp = str;
  while((bp=strchr(bp,'\001'))!=NULL) {*bp++=' ';}

  if (!m_file.mm_flg) fseek(sfile,m_file.s_pos_arr[libpos],0);

  bl_lib_pos = m_file.s_pos_arr[libpos];
}

unsigned int bl2_uint4_cvt(unsigned int val)
{
  unsigned int res;
#ifdef IS_BIG_ENDIAN
  return val;
#else /* it better be LITTLE_ENDIAN */
  res = ((val&255)*256)+ ((val>>8)&255);
  res = (res<<16) + (((val>>16)&255)*256) + ((val>>24)&255);
  return res;
#endif
}  

unsigned int bl2_long4_cvt(long val)
{
  int val4;
  unsigned int res;
#ifdef IS_BIG_ENDIAN
  val4 = val;
  return val4;
#else /* it better be LITTLE_ENDIAN */
  res = ((val&255)*256)+ ((val>>8)&255);
  res = (res<<16) + (((val>>16)&255)*256) + ((val>>24)&255);
  return res;
#endif
}  

long bl2_long8_cvt(long val)
{
  long res;
#ifdef IS_BIG_ENDIAN
  return val;
#else /* it better be LITTLE_ENDIAN */
  res = ((val&255)*256)+ ((val>>8)&255);
  res = (res<<16) + (((val>>16)&255)*256) + ((val>>24)&255);
#ifdef BIG_LIB64
  res = (res<<16) + (((val>>32)&255)*256) + ((val>>40)&255);
  res = (res<<16) + (((val>>48)&255)*256) + ((val>>56)&255);
#else
  fprintf(stderr,"Cannot use bl2_long8_cvt without 64 bit longs\n");
  exit(1);
#endif
  return res;
#endif
}  

void src_uint8_read(FILE *fd, unsigned long *val)
{
#ifdef IS_BIG_ENDIAN
  fread((char *)val,sizeof(long),(size_t)1,fd);
#else
  unsigned char b[8];

  fread((char *)&b[0],(size_t)1,(size_t)8,fd);
  *val = 0;
  *val = (long)((((((long)((long)(b[0]<<8)+(long)b[1]<<8)+(long)b[2]<<8)
		   +((long)b[3]<<8))+((long)b[4]<<8))+((long)b[5]<<8))
		   +((long)b[6]<<8))+(long)b[7];
#endif
}

void src_int4_read(FILE *fd,  int *val)
{
#ifdef IS_BIG_ENDIAN
  fread((char *)val,(size_t)4,(size_t)1,fd);
#else
  unsigned char b[4];

  fread((char *)&b[0],(size_t)1,(size_t)4,fd);
  *val = 0;
  *val = (int)((int)((int)(b[0]<<8)+(int)b[1]<<8)+(int)b[2]<<8)
	  +(int)b[3];
#endif
}

void src_long4_read(FILE *fd,  long *valp)
{
  int val4;
#ifdef IS_BIG_ENDIAN
  fread(&val4,(size_t)4,(size_t)1,fd);
  *valp = val4;
#else
  unsigned char b[4];

  fread((char *)&b[0],(size_t)1,(size_t)4,fd);
  val4 = 0;
  val4 = (int)((int)((int)(b[0]<<8)+(int)b[1]<<8)+(int)b[2]<<8)
	  +(int)b[3];
  *valp = val4;
#endif
}

void src_ulong4_read(FILE *fd,  unsigned long *valp)
{
#ifdef IS_BIG_ENDIAN
  unsigned int val4;
  fread(&val4,(size_t)4,(size_t)1,fd);
  *valp = val4;
#else
  unsigned long val4;
  unsigned char b[4];

  fread((char *)&b[0],(size_t)1,(size_t)4,fd);
  val4 = 0;
  val4 = (int)((int)((int)(b[0]<<8)+(int)b[1]<<8)+(int)b[2]<<8)
	  +(int)b[3];
  *valp = val4;
#endif
}

void src_uint4_read(FILE *fd,  unsigned int *valp)
{
#ifdef IS_BIG_ENDIAN
  fread(valp,(size_t)4,(size_t)1,fd);
#else
  unsigned char b[4];

  fread((char *)&b[0],(size_t)1,(size_t)4,fd);
  *valp = 0;
  *valp = (int)((int)((int)(b[0]<<8)+(int)b[1]<<8)+(int)b[2]<<8)
	  +(int)b[3];
#endif
}

void src_long8_read(FILE *fd,  long *val)
{
#ifdef IS_BIG_ENDIAN
  fread((void *)val,(size_t)8,(size_t)1,fd);
#else
  unsigned char b[8];

  fread((char *)&b[0],(size_t)1,(size_t)8,fd);
  *val = 0;
  *val = (long)((((((long)((long)(b[0]<<8)+(long)b[1]<<8)+(long)b[2]<<8)
		  +(long)b[3]<<8)+(long)b[4]<<8)+(long)b[5]<<8)
		+(long)b[6]<<8)+(long)b[7];
#endif
}

void src_char_read(FILE *fd, char *val)
{
  fread(val,(size_t)1,(size_t)1,fd);
}

void src_fstr_read(FILE *fd, char *val,  int slen)
{
  fread(val,(size_t)slen,(size_t)1,fd);
}

void
newname(char *nname, char *oname, char *suff, int maxn)
{
  strncpy(nname,oname,maxn-1);
  strncat(nname,".",1);
  strncat(nname,suff,maxn-strlen(nname));
}
